-- Created by Giseldah

local p = _G; --p stands for package

import "Giseldah.CalcStat.CalcStatFull"; -- import from Lotro plugin environment

-- *********************** Start CalcStat Expressions *************************

local TRANS_EXPRESSION = 1;
local TRANS_EXPRWIP = 2;
local TRANS_ERRCODE = 3;
local TRANS_OPERATIONS = 4;
local TRANS_OPERANDS = 5;
local TRANS_VARIABLES = 6;

local OPERAND_SOURCE = 1; -- source expression for operand (should be unique)
local OPERAND_VALUE = 2; -- current value of operand
local OPERAND_OPERATION = 3; -- source operation which gives value to this operand
local OPERAND_DEPENDSON = 4; -- bit flags: collection of variables (indexes) on which operand depends for it's value

local OPERAND_OPERATION_GETVAR = -1; -- custom operation: retrieve variable value

local VAR_NAME = 1;
local VAR_VALUE = 2;
local VAR_OPERANDS = 3; -- list of operands which depend on this variable

local OPERATION_DEFINITION = 1; -- index into aOperationDefs table
local OPERATION_OPERANDS = 2; -- parameter operand indexes into TRANS_OPERANDS
local OPERATION_RESULTOP = 3; -- result operand index into TRANS_OPERANDS

local OPERATION_FORMAT_FUNCTION = 1; -- like functionname(op1,op2,op3,etc) - needs an unspecified number of operands
local OPERATION_FORMAT_UNARY = 2; -- like: operator op2 - always needs 1 operand
local OPERATION_FORMAT_BINARY = 3; -- like: op1 operator op2 - always needs 2 operands

-- operation definition Id's (not to be confused with indexes into aOperationDefs)
local UNARYOP_LOGICNEG = 1
local UNARYOP_NEGATE = 2;
local UNARYOP_POSITIVE = 3;
local BINARYOP_CONCAT = 4;
local BINARYOP_OR = 5;
local BINARYOP_AND = 6;
local BINARYOP_GREQTH = 7;
local BINARYOP_SMEQTH = 8;
local BINARYOP_EQUALS = 9;
local BINARYOP_NOTEQUALS = 10;
local BINARYOP_GRTH = 11;
local BINARYOP_SMTH = 12;
local BINARYOP_SUBSTRACT = 13;
local BINARYOP_ADD = 14;
local BINARYOP_MOD = 15;
local BINARYOP_MULTIPLY = 16;
local BINARYOP_DIVIDE = 17;
local BINARYOP_POWER = 18;
local FUNCTIONOP_IIF = 19;
local FUNCTIONOP_CALCSTAT = 20;
local FUNCTIONOP_ABS = 21;
local FUNCTIONOP_CEIL = 22;
local FUNCTIONOP_FLOOR = 23;
local FUNCTIONOP_MAX = 24;
local FUNCTIONOP_MIN = 25;
local FUNCTIONOP_LOG10 = 26;
local FUNCTIONOP_SQRT = 27;
local FUNCTIONOP_AFD = 28;
local FUNCTIONOP_FORMAT = 29;
local FUNCTIONOP_SWITCH = 30;
local FUNCTIONOP_SESEL = 31;
local FUNCTIONOP_WITH = 32;

-- operators and function names to detect (always in uppercase)
local OPERATOR_LOGICNEG = "NOT";
local OPERATOR_NEGATE = "-";
local OPERATOR_POSITIVE = "+";
local OPERATOR_CONCAT = "..";
local OPERATOR_OR = "OR";
local OPERATOR_AND = "AND";
local OPERATOR_GREQTH = ">=";
local OPERATOR_SMEQTH = "<=";
local OPERATOR_EQUALS = "==";
local OPERATOR_NOTEQUALS = "~=";
local OPERATOR_GRTH = ">";
local OPERATOR_SMTH = "<";
local OPERATOR_SUBSTRACT = "-";
local OPERATOR_ADD = "+";
local OPERATOR_MOD = "%";
local OPERATOR_MULTIPLY = "*";
local OPERATOR_DIVIDE = "/";
local OPERATOR_POWER = "^";
local FUNCNAME_IIF = "IIF";
local FUNCNAME_CALCSTAT = "CALCSTAT";
local FUNCNAME_ABS = "ABS";
local FUNCNAME_CEIL = "CEIL";
local FUNCNAME_FLOOR = "FLOOR";
local FUNCNAME_MAX = "MAX";
local FUNCNAME_MIN = "MIN";
local FUNCNAME_LOG10 = "LOG10";
local FUNCNAME_SQRT = "SQRT";
local FUNCNAME_AFD = "AFD";
local FUNCNAME_FORMAT = "FORMAT";
local FUNCNAME_SWITCH = "SWITCH";
local FUNCNAME_SESEL = "SESEL";
local FUNCNAME_WITH = "WITH";

local OPERATIONDEF_ID = 1;
local OPERATIONDEF_FORMAT = 2;
local OPERATIONDEF_FUNCNAME = 3;
local OPERATIONDEF_OPERATOR = 3;

local aOperationDefs = {
						{UNARYOP_LOGICNEG,OPERATION_FORMAT_UNARY,OPERATOR_LOGICNEG},
						{UNARYOP_NEGATE,OPERATION_FORMAT_UNARY,OPERATOR_NEGATE},
						{UNARYOP_POSITIVE,OPERATION_FORMAT_UNARY,OPERATOR_POSITIVE},
						{BINARYOP_CONCAT,OPERATION_FORMAT_BINARY,OPERATOR_CONCAT},
						{BINARYOP_OR,OPERATION_FORMAT_BINARY,OPERATOR_OR},
						{BINARYOP_AND,OPERATION_FORMAT_BINARY,OPERATOR_AND},
						{BINARYOP_GREQTH,OPERATION_FORMAT_BINARY,OPERATOR_GREQTH},
						{BINARYOP_SMEQTH,OPERATION_FORMAT_BINARY,OPERATOR_SMEQTH},
						{BINARYOP_EQUALS,OPERATION_FORMAT_BINARY,OPERATOR_EQUALS},
						{BINARYOP_NOTEQUALS,OPERATION_FORMAT_BINARY,OPERATOR_NOTEQUALS},
						{BINARYOP_GRTH,OPERATION_FORMAT_BINARY,OPERATOR_GRTH},
						{BINARYOP_SMTH,OPERATION_FORMAT_BINARY,OPERATOR_SMTH},
						{BINARYOP_SUBSTRACT,OPERATION_FORMAT_BINARY,OPERATOR_SUBSTRACT},
						{BINARYOP_ADD,OPERATION_FORMAT_BINARY,OPERATOR_ADD},
						{BINARYOP_MOD,OPERATION_FORMAT_BINARY,OPERATOR_MOD},
						{BINARYOP_MULTIPLY,OPERATION_FORMAT_BINARY,OPERATOR_MULTIPLY},
						{BINARYOP_DIVIDE,OPERATION_FORMAT_BINARY,OPERATOR_DIVIDE},
						{BINARYOP_POWER,OPERATION_FORMAT_BINARY,OPERATOR_POWER},
						{FUNCTIONOP_IIF,OPERATION_FORMAT_FUNCTION,FUNCNAME_IIF},
						{FUNCTIONOP_CALCSTAT,OPERATION_FORMAT_FUNCTION,FUNCNAME_CALCSTAT},
						{FUNCTIONOP_ABS,OPERATION_FORMAT_FUNCTION,FUNCNAME_ABS},
						{FUNCTIONOP_CEIL,OPERATION_FORMAT_FUNCTION,FUNCNAME_CEIL},
						{FUNCTIONOP_FLOOR,OPERATION_FORMAT_FUNCTION,FUNCNAME_FLOOR},
						{FUNCTIONOP_MAX,OPERATION_FORMAT_FUNCTION,FUNCNAME_MAX},
						{FUNCTIONOP_MIN,OPERATION_FORMAT_FUNCTION,FUNCNAME_MIN},
						{FUNCTIONOP_LOG10,OPERATION_FORMAT_FUNCTION,FUNCNAME_LOG10},
						{FUNCTIONOP_SQRT,OPERATION_FORMAT_FUNCTION,FUNCNAME_SQRT},
						{FUNCTIONOP_AFD,OPERATION_FORMAT_FUNCTION,FUNCNAME_AFD},
						{FUNCTIONOP_FORMAT,OPERATION_FORMAT_FUNCTION,FUNCNAME_FORMAT},
						{FUNCTIONOP_SWITCH,OPERATION_FORMAT_FUNCTION,FUNCNAME_SWITCH},
						{FUNCTIONOP_SESEL,OPERATION_FORMAT_FUNCTION,FUNCNAME_SESEL},
						{FUNCTIONOP_WITH,OPERATION_FORMAT_FUNCTION,FUNCNAME_WITH}
						};

local OPPRIOGRP_OPERATIONDEF_IDS = 1;
local OPPRIOGRP_SCANDIRECTION = 2;
local OPPRIOGRP_SCANFORMAT = 3;

local SCANDIRECTION_LEFTRIGHT = 1;
local SCANDIRECTION_RIGHTLEFT = 2;

local ERR_TOOMANYVARIABLES = 1;
local ERR_EXPRESSIONSYNTAX = 2;
local ERR_INVALIDOPCOUNT = 3;
local ERR_INVALIDOPTYPE = 4;

local aErrMsgs = {
					"Too many variables used",
					"Syntax error",
					"Invalid number of operands",
					"Invalid operand type",
					};

-- defines priority process order and scan direction for grouped binary and unary operations
-- same as Lua operation order, highest priority first
-- you can't mix unary and binary operations in the same group: it's either all unary or all binary
local aOperationPriorityGroups = {
									{{BINARYOP_POWER},SCANDIRECTION_RIGHTLEFT,OPERATION_FORMAT_BINARY},
									{{UNARYOP_LOGICNEG,UNARYOP_NEGATE,UNARYOP_POSITIVE},SCANDIRECTION_RIGHTLEFT,OPERATION_FORMAT_UNARY},
									{{BINARYOP_MULTIPLY,BINARYOP_DIVIDE,BINARYOP_MOD},SCANDIRECTION_LEFTRIGHT,OPERATION_FORMAT_BINARY},
									{{BINARYOP_ADD,BINARYOP_SUBSTRACT},SCANDIRECTION_LEFTRIGHT,OPERATION_FORMAT_BINARY},
									{{BINARYOP_CONCAT},SCANDIRECTION_RIGHTLEFT,OPERATION_FORMAT_BINARY},
									{{BINARYOP_SMTH,BINARYOP_GRTH,BINARYOP_SMEQTH,BINARYOP_GREQTH,BINARYOP_NOTEQUALS,BINARYOP_EQUALS},SCANDIRECTION_LEFTRIGHT,OPERATION_FORMAT_BINARY},
									{{BINARYOP_AND},SCANDIRECTION_LEFTRIGHT,OPERATION_FORMAT_BINARY},
									{{BINARYOP_OR},SCANDIRECTION_LEFTRIGHT,OPERATION_FORMAT_BINARY}
									};

-- native bitwise operations sadly only from Lua 5.2 up
-- bitwise or
local function BitOR(a,b)
	local p,c = 1,0
	while a+b > 0 do
		local ra,rb = a%2,b%2;
		if ra+rb > 0 then
			c = c+p;
		end
		a,b,p = (a-ra)/2, (b-rb)/2, p*2;
	end
	return c;
end

-- returns operand number from operand string like 9 from "[9]"
local function OPStonumber(sOPS)
	local sOPn = sOPS:match("^%s*%[(%d+)%]%s*$");
	if sOPn == nil then
		return nil;
	else
		return tonumber(sOPn);
	end
end

-- returns operand string from operand number/string like "[9]" from 9
local function OPNtostring(xOPN)
	if xOPN == nil then
		return nil;
	else
		return "["..xOPN.."]";
	end
end

-- returns index into TRANS_OPERANDS for a given sSource
local function FindOperandBySource(aTranslation,sSource)
	for i, o in pairs(aTranslation[TRANS_OPERANDS]) do
		if o[OPERAND_SOURCE] == sSource then
			return i;
		end
	end
	return nil;
end

-- checks if a variable is in the variable table and returns its index if so
local function FindVariableByName(aTranslation,sName)
	for i, v in pairs(aTranslation[TRANS_VARIABLES]) do
		if v[VAR_NAME] == sName then
			return i;
		end
	end
	return nil;
end

-- returns an index into the Operation Definition table (aOperationDefs) given an Id
local function GetOperationDefById(nDefId)
	for i, d in pairs(aOperationDefs) do
		if d[OPERATIONDEF_ID] == nDefId then
			return i;
		end
	end
	return nil;
end

-- returns an index into the Operation Definition table (aOperationDefs) given a function name
local function GetOperationDefByFunctionName(sName)
	for i, d in pairs(aOperationDefs) do
		if d[OPERATIONDEF_FORMAT] == OPERATION_FORMAT_FUNCTION and d[OPERATIONDEF_FUNCNAME] == sName then
			return i;
		end
	end
	return nil;
end

-- register operand to variables as dependable
local function RegisterOperandWithDependsOn(aTranslation,nDependsOn,nOp)
	if nDependsOn == 0 then
		return;
	end
	local nVar, nVarCount, nCheckBits, nBit;
	nVarCount = #aTranslation[TRANS_VARIABLES];
	nCheckBits = nDependsOn;
	for nVar = 1, nVarCount do
		nBit = nCheckBits % 2;
		if nBit > 0 then
			-- operand depends on this variable
			table.insert(aTranslation[TRANS_VARIABLES][nVar][VAR_OPERANDS],nOp); -- register operand number in variable as dependable
		end
		nCheckBits = (nCheckBits-nBit) / 2;
	end
end

-- detects a boolean value in a text, creates an operand for it and returns operand index
-- if the value already exists then it returns the existing operand index
-- pattern matches FALSE, TRUE
local function CreateBooleanConstant(aTranslation,sBoolean)
    local sSource = sBoolean:match("%s*(FALSE)%s*");
	if sSource == nil then
		sSource = sBoolean:match("%s*(TRUE)%s*");
	end
	local nOp = FindOperandBySource(aTranslation,sSource);
	if nOp == nil then
		-- add new operand if not exists yet
		nOp = #aTranslation[TRANS_OPERANDS]+1;
		local aOperand = {};
		aOperand[OPERAND_SOURCE] = sSource;
		aOperand[OPERAND_VALUE] = (sSource == "TRUE"); -- assign value to this static operand
		aOperand[OPERAND_OPERATION] = nil; -- value is not created by an operation
		aOperand[OPERAND_DEPENDSON] = 0; -- doesn't depend on any either down the chain (isn't any)
		aTranslation[TRANS_OPERANDS][nOp] = aOperand;
	end
	return OPNtostring(nOp);
end

-- detects a number value in a text, creates an operand for it and returns operand index
-- if the value already exists then it returns the existing operand index
-- pattern matches the minimal 9 and 9.9 or more digits
local function CreateNumberConstant(aTranslation,sNumber)
    local sSource = sNumber:match("%s*(%d+%.%d+)%s*");
	if sSource == nil then
		sSource = sNumber:match("%s*(%d+)%s*");
	end
	local nOp = FindOperandBySource(aTranslation,sSource);
	if nOp == nil then
		-- add new operand if not exists yet
		nOp = #aTranslation[TRANS_OPERANDS]+1;
		local aOperand = {};
		aOperand[OPERAND_SOURCE] = sSource;
		aOperand[OPERAND_VALUE] = tonumber(sSource); -- assign value to this static operand
		aOperand[OPERAND_OPERATION] = nil; -- value is not created by an operation
		aOperand[OPERAND_DEPENDSON] = 0; -- doesn't depend on any either down the chain (isn't any)
		aTranslation[TRANS_OPERANDS][nOp] = aOperand;
	end
	return OPNtostring(nOp);
end

-- detects a string value in a text, creates an operand for it and returns operand index
-- if the value already exists then it returns the existing operand index
-- pattern matches characters between ""
local function CreateStringConstant(aTranslation,sString)
    local sSource = sString:match("%s*(\".-\")%s*");
	local nOp = FindOperandBySource(aTranslation,sSource);
	if nOp == nil then
		-- add new operand if not exists yet
		nOp = #aTranslation[TRANS_OPERANDS]+1;
		local aOperand = {};
		aOperand[OPERAND_SOURCE] = sSource;
		aOperand[OPERAND_VALUE] = sSource:sub(2,-2); -- assign value to this static operand
		aOperand[OPERAND_OPERATION] = nil; -- value is not created by an operation
		aOperand[OPERAND_DEPENDSON] = 0; -- doesn't depend on any either down the chain (isn't any)
		aTranslation[TRANS_OPERANDS][nOp] = aOperand;
	end
	return OPNtostring(nOp);
end

-- extracts string operands out of the expression and replaces them with the operand index indicator '[n]'
local function CreateStringConstants(aTranslation)
	aTranslation[TRANS_EXPRWIP] = aTranslation[TRANS_EXPRWIP]:gsub("%s*\".-\"%s*",function(sString) return CreateStringConstant(aTranslation,sString) end);
end

-- detects a variable in a text, creates an operand for it and returns operand index
-- if an operand for the variable already exists then it returns the existing operand index
-- pattern matches the minimal $A (or with more letters)
-- operand index is registered with the variable's entry in the variable table
-- dependson flag ensures that every operand that uses this operand will also depend on the variable and be registered as such
local function CreateVariableOperand(aTranslation,sVar)
    local sSource = sVar:match("%s*(%$%a+)%s*");
	local nVar = FindVariableByName(aTranslation,sSource);
	if nVar == nil then
		-- add new 'on the fly' variable. inside With function?
		nVar = #aTranslation[TRANS_VARIABLES]+1; -- maximum 31? (32 bits minus sign bit)
		if nVar > 31 then
			aTranslation[TRANS_ERRCODE] = ERR_TOOMANYVARIABLES;
			return nil;
		end
		local aVar = {};
		aVar[VAR_NAME] = sSource;
		aVar[VAR_VALUE] = nil; -- initial value
		aVar[VAR_OPERANDS] = {}; -- operands registration table
		aTranslation[TRANS_VARIABLES][nVar] = aVar;
	end
	local nOp = FindOperandBySource(aTranslation,sSource);
	if nOp == nil then
		-- add new operand if not exists yet
		nOp = #aTranslation[TRANS_OPERANDS]+1;
		local aOperand = {};
		aOperand[OPERAND_SOURCE] = sSource;
		aOperand[OPERAND_VALUE] = nil;
		aOperand[OPERAND_OPERATION] = OPERAND_OPERATION_GETVAR; -- value needs to be picked up from variable
		local nDependsOn = 2^(nVar-1);
		aOperand[OPERAND_DEPENDSON] = nDependsOn; -- depends on variable flag
		aTranslation[TRANS_OPERANDS][nOp] = aOperand;
		RegisterOperandWithDependsOn(aTranslation,nDependsOn,nOp); -- register operand with variable(s) it depends on
	end
	return OPNtostring(nOp);
end

-- replaces the short stat notations for the calcstat function, like @STATNAME# etc
-- the short notations will be replaced by the normal CALCSTAT("STATNAME",$L,$N) or more complex
-- while we are at it, any variable operand use for $L/$N implicated or constant number operands detected, are being processed on the fly
local function ReplaceCalcStatShorts(aTranslation)
	local sReplExpr = aTranslation[TRANS_EXPRWIP];
	local aPatterns = {
						{"%s*@(%w+)%s*(%b())%s*(#)%s*(%b())%s*",true,true,true}, -- @a ( $L ) # ( $N ) 
						{"%s*@(%w+)%s*(%b())%s*(#)%s*(%d+%.%d+)%s*",true,true,true}, -- @a ( $L ) # 0.1 
						{"%s*@(%w+)%s*(%b())%s*(#)%s*(%d+)%s*",true,true,true}, -- @a ( $L ) # 1 
						{"%s*@(%w+)%s*(%b())%s*(#)%s*",true,true,false}, -- @a ( $L ) # 
						{"%s*@(%w+)%s*(%b())%s*",true,false,false}, -- @a ( $L ) 
						{"%s*@(%w+)%s*(#)%s*(%b())%s*",false,true,true}, -- @a # ( $N ) 
						{"%s*@(%w+)%s*(#)%s*(%d+%.%d+)%s*",false,true,true}, -- @a # 0.1 
						{"%s*@(%w+)%s*(#)%s*(%d+)%s*",false,true,true}, -- @a # 1 
						{"%s*@(%w+)%s*(#)%s*",false,true,false}, -- @a # 
						{"%s*@(%w+)%s*",false,false,false}, -- @a  
					};
	local nShortStart, nShortEnd, sSN, sL, sHash, sN;
	local nSNLevelStart, nSNLevelEnd, sSNLevel;
	for _, aPattern in pairs(aPatterns) do
		repeat
			if aPattern[2] then
				-- contains level
				nShortStart, nShortEnd, sSN, sL, sHash, sN = sReplExpr:find(aPattern[1]);
			else
				nShortStart, nShortEnd, sSN, sHash, sN = sReplExpr:find(aPattern[1]);
				sL = nil;
			end
			if nShortStart ~= nil then
				if aPattern[2] then
					sL = sL:match("^%(%s*(.-)%s*%)$"); -- remove brackets
				else
					-- test for level in sSN
					nSNLevelStart, nSNLevelEnd, sSNLevel = sSN:find("(%d+)$");
					if nSNLevelStart == nil then
						sL = CreateVariableOperand(aTranslation,"$L"); -- create level variable (or get if already exists)
						if sL == nil then
							break;
						end
					else
						sSN = sSN:sub(1,nSNLevelStart-1); -- first part minus level
						sL = CreateNumberConstant(aTranslation,sSNLevel); -- transform level to number constant
					end
				end
				sSN = CreateStringConstant(aTranslation,"\""..sSN.."\""); -- transform statname to string constant
				if aPattern[3] then
					if aPattern[4] then
						if sN:sub(1,1) == "(" then
							sN = sN:match("^%(%s*(.-)%s*%)$"); -- remove brackets if expression
						else
							sN = CreateNumberConstant(aTranslation,sN); -- transform value for $N to number constant
						end
					else
						sN = CreateVariableOperand(aTranslation,"$N"); -- create N variable (or get if already exists);
						if sN == nil then
							break;
						end
					end
					sReplExpr = sReplExpr:sub(1,nShortStart-1).." CALCSTAT("..sSN..","..sL..","..sN..")"..sReplExpr:sub(nShortEnd+1);
				else
					-- sN = "nil";
					sReplExpr = sReplExpr:sub(1,nShortStart-1).." CALCSTAT("..sSN..","..sL..")"..sReplExpr:sub(nShortEnd+1); -- just leave out $N|$C parameter if default
				end
			end
		until nShortStart == nil
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			break;
		end
	end
	aTranslation[TRANS_EXPRWIP] = sReplExpr;
end

-- removes all variable references from the expression and replaces them with operand identifiers (indexes in operand table)
-- for example "blah blah $TY lefhoihfoe $HIG fslhf" will look like "blah blah [1] lefhoihfoe [2] fslhf"
-- operands with indexes 1 and 2 have then been created in the operand table
local function CreateVariableOperands(aTranslation)
	aTranslation[TRANS_EXPRWIP] = aTranslation[TRANS_EXPRWIP]:gsub("%s*%$%a+%s*",function(sVar) if aTranslation[TRANS_ERRCODE] == 0 then return CreateVariableOperand(aTranslation,sVar) else return sVar end end);
end

-- removes number and boolean constant values from an expression text and replaces them with operand identifiers (indexes in operand table)
-- for example " TRUE , 1.55 " will look something like "[1],[2]"
-- operands with indexes 1 and 2 have then been created in the operand table
local function CreateConstantOperands(aTranslation,sExpression)
	local sReplExpr = sExpression;
	-- booleans
	sReplExpr = sReplExpr:gsub("%s*FALSE%s*",function(sBoolean) return CreateBooleanConstant(aTranslation,sBoolean) end);
	sReplExpr = sReplExpr:gsub("%s*TRUE%s*",function(sBoolean) return CreateBooleanConstant(aTranslation,sBoolean) end);
	-- numbers
	sReplExpr = sReplExpr:gsub("%s*%d+%.%d+%s*",function(sNumber) return CreateNumberConstant(aTranslation,sNumber) end)
	-- problem: numbers without dot in operand references [n]
	sReplExpr = sReplExpr:gsub("(%s*)([^%d]?)(%d+)([^%d]?)(%s*)",function(sStart,sTest1,sNumber,sTest2,sEnd) if not (sTest1 == "[" and sTest2 == "]") then return sStart..sTest1..CreateNumberConstant(aTranslation,sNumber)..sTest2..sEnd else return sStart..sTest1..sNumber..sTest2..sEnd end end); -- just %s*%d+%s* would match n in [n]
	return sReplExpr;
end

local DecomposeExpression;

-- splits an expression into (parameter)parts, based on a splitting token
-- calls a supplied function for each detected part
-- tokens in nested expressions between () are ignored
local function DecomposeParameters(sParams,sSplitToken,fDecomposeParameter)
	local nLevel = 0; -- to keep track of nested ()
	local nCharIndex;
	local nParamIndex = 1; -- points to start position of parameter
	local nLastIndex = sParams:len();
	local sChar;
	local nParam = 0;
	for nCharIndex = 1, nLastIndex do
		sChar = sParams:sub(nCharIndex,nCharIndex);
		if sChar == "(" then
			nLevel = nLevel+1
		elseif sChar == ")" then
			nLevel = nLevel-1
		end
		if nLevel == 0 then
			if sChar == sSplitToken then
				-- we have a parameter when we detect a splitting token at level 0
				nParam = nParam+1;
				fDecomposeParameter(sParams:sub(nParamIndex,nCharIndex-1),nParam,false);
				nParamIndex = nCharIndex+1;
			end
			if nCharIndex == nLastIndex then
				-- last parameter at the end
				fDecomposeParameter(sParams:sub(nParamIndex),nParam+1,true);
			end
		end
	end
end

-- sParams = param1,param2,etc
-- returns an array with operand numbers, containing parameter expression results
local function DecomposeFunctionParameters(aTranslation,sParams)
	-- extract parameters
	local aParamOps = {};

	local DecomposeFunctionParameter = function(sParam,nParam,bIsLastParam)
		if sParam == "" then
			aParamOps[#aParamOps+1] = 999999999; -- empty parameter, might be used for default behaviours
		else
			aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sParam));
		end
	end
	
	DecomposeParameters(sParams,",",DecomposeFunctionParameter);
	
	return aParamOps;
end

-- transform CASE>=compexpr1:CASE<=compexpr2:..etc:resultexpr
-- expand to: (testexpr>=compexpr1)OR(testexpr<=compexpr2)OR..etc
local function DecomposeCaseConditions(aTranslation,sParams,sTestOp,aParamOps)
	local sCombinedExpr = "";

	local DecomposeCaseCondition = function(sParam,nParam,bIsLastParam)
		if bIsLastParam then
			-- add combined logical expression
			aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sCombinedExpr));
			-- add last: should be result expression
			aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sParam));
		else
			-- construct combined logical expression
			local sCompareExpr = sParam:match("^%s*CASE%s*(.-)%s*$");
			if sCombinedExpr ~= "" then
				sCombinedExpr = sCombinedExpr.."OR";
			end
			sCombinedExpr = sCombinedExpr..DecomposeExpression(aTranslation,sTestOp..sCompareExpr);
		end
	end
	
	DecomposeParameters(sParams,":",DecomposeCaseCondition);
end

-- SWITCH(testexpr,CASE>=compexpr1:CASE<=compexpr2:resultexpr1,DEFAULT:resultexpr2)
-- expand to: SWITCH((testexpr>=compexpr1)OR(testexpr<=compexpr2),resultexpr1,True,resultexpr2)
-- parameter results should be like bBoolean1,xValue1,bBoolean2,xValue2,bBoolean3,xValue3,etc
local function DecomposeSwitchParameters(aTranslation,sParams)
	-- extract parameters
	local aParamOps = {};
	local sTestOp;
	
	local DecomposeSwitchParameter = function(sParam,nParam,bIsLastParam)
		if nParam == 1 then
			-- should be test expression
			sTestOp = DecomposeExpression(aTranslation,sParam);
		else
			-- test for CASE or DEFAULT at the start
			if sParam:match("^%s*CASE") ~= nil then
				-- CASE parameter, possibly multiple conditions
				-- CASE>=compexpr1:CASE<=compexpr2:resultexpr1
				DecomposeCaseConditions(aTranslation,sParam,sTestOp,aParamOps);
			elseif sParam:match("^%s*DEFAULT") ~= nil then
				-- default parameter
				-- DEFAULT:resultexpr2
				local sDefaultExpr = sParam:match("^%s*DEFAULT%s*:%s*(.-)%s*$");
				-- add true parameter operand
				aParamOps[#aParamOps+1] = OPStonumber(CreateBooleanConstant(aTranslation,"TRUE"));
				-- add default result parameter operand
				aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sDefaultExpr));
			end
		end
	end
	
	DecomposeParameters(sParams,",",DecomposeSwitchParameter);
	
	return aParamOps;
end

-- var1=varexpr1, var2=varexpr2, etc.
-- expand to: var1, varexpr1, var2, varexpr2, etc.
local function DecomposeWithVarAssignments(aTranslation,sParams,aParamOps)
	local DecomposeWithVarAssignment = function(sParam,nParam,bIsLastParam)
		local sVar, sVarExpr = sParam:match("^%s*(%[%d+%])%s*=%s*(.-)%s*$"); -- variable should already be an operand
		if sVar ~= nil and sVarExpr ~= nil then
			-- add variable operand
			aParamOps[#aParamOps+1] = OPStonumber(sVar);
			-- add assigned expression
			aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sVarExpr));
		end
	end
	
	DecomposeParameters(sParams,",",DecomposeWithVarAssignment);
end

-- WITH(var1=varexpr1, var2=varexpr2, etc. : resultexpr)
-- expand to: WITH(var1, varexpr1, var2, varexpr2, etc. , resultexpr)
-- parameter results should be like a pair for each variable assignment plus one result operand at the end
local function DecomposeWithParameters(aTranslation,sParams)
	-- extract parameters
	local aParamOps = {};
	
	local DecomposeWithParameter = function(sParam,nParam,bIsLastParam)
		if nParam == 1 then
			-- should be variable assignments
			DecomposeWithVarAssignments(aTranslation,sParam,aParamOps);
		elseif bIsLastParam then
			-- should be the result expression
			-- add result parameter operand
			aParamOps[#aParamOps+1] = OPStonumber(DecomposeExpression(aTranslation,sParam));
		end
	end
	
	DecomposeParameters(sParams,":",DecomposeWithParameter);
	
	return aParamOps;
end

local function GetSourceParameterOperands(aParamOps)
	local sSource = "";
	for i, nOp in pairs(aParamOps) do
		if sSource ~= "" then
			sSource = sSource..",";
		end
		if nOp ~= nil then
			sSource = sSource..OPNtostring(nOp);
		end
	end
	return sSource;
end

local ExecuteOperation;
local function CreateOperation(aTranslation,sSource,nOperationDef,aParamOps)
	-- test if a result not yet exists for this source
	local nResultOp = FindOperandBySource(aTranslation,sSource);
	if nResultOp == nil then
		-- compose combined dependson for these operands
		local nDependsOn = 0;
		if aParamOps ~= nil then
			local nOp;
			for _, nOp in pairs(aParamOps) do
				if nOp ~= nil then
					nDependsOn = BitOR(nDependsOn,aTranslation[TRANS_OPERANDS][nOp][OPERAND_DEPENDSON]);
				end
			end
		end

		-- create result operand
		nResultOp = #aTranslation[TRANS_OPERANDS]+1;
		local aOperand = {};
		aOperand[OPERAND_SOURCE] = sSource;
		aOperand[OPERAND_VALUE] = nil;
		aOperand[OPERAND_OPERATION] = nil;
		aOperand[OPERAND_DEPENDSON] = nDependsOn;
		aTranslation[TRANS_OPERANDS][nResultOp] = aOperand;
		RegisterOperandWithDependsOn(aTranslation,nDependsOn,nResultOp); -- register operand with variable(s) it depends on

		-- create the operation
		local aOperation = {};
		aOperation[OPERATION_DEFINITION] = nOperationDef;
		aOperation[OPERATION_OPERANDS] = aParamOps;
		aOperation[OPERATION_RESULTOP] = nResultOp;

		if nDependsOn == 0 then
			-- does not depend on any variable, so operand values should be all constants
			-- calculate constant value now and don't store operation after
			ExecuteOperation(aTranslation,aOperation);
		else
			-- depends on variable(s), so operation needs to be stored, but not executed now
			local nResultOperation = #aTranslation[TRANS_OPERATIONS]+1;
			aTranslation[TRANS_OPERANDS][nResultOp][OPERAND_OPERATION] = nResultOperation; -- register operation in operand
			aTranslation[TRANS_OPERATIONS][nResultOperation] = aOperation; -- store in operation table
		end
	end
	return nResultOp;
end

-- sName = name of function like CALCSTAT, LINFMOD, etc.
-- sParams = param1,param2,etc (normally)
-- parameters need to be extracted and each treated like an expression (which should return a result in an operand)
-- an operation for this function needs to be created, which takes a list of operands (the parameters) and gives a result in an operand
local function DecomposeFunction(aTranslation,sName,sParams)
	-- get function definition
	local nOperationDef = GetOperationDefByFunctionName(sName);
	if nOperationDef == nil then
		-- unknown function
		return nil;
	end

	-- get an array with parameter operand numbers
	local aParamOps;
	if aOperationDefs[nOperationDef][OPERATIONDEF_ID] == FUNCTIONOP_SWITCH then
		aParamOps = DecomposeSwitchParameters(aTranslation,sParams); -- Switch($L, case<=10: case>=100: 25, default: 13)
	elseif aOperationDefs[nOperationDef][OPERATIONDEF_ID] == FUNCTIONOP_WITH then
		aParamOps = DecomposeWithParameters(aTranslation,sParams); -- With($A=@Agility#2, $B=20: IIf($A>$B,$A,$B))
	else
		aParamOps = DecomposeFunctionParameters(aTranslation,sParams); -- sName(operand1, operand2, etc.)
	end

	-- compose funcname([n],[n],etc) as source identifier
	local sSource = sName.."("..GetSourceParameterOperands(aParamOps)..")";

	-- create an operation for this function (returns result operand number)
	local nResultOp = CreateOperation(aTranslation,sSource,nOperationDef,aParamOps);

	return OPNtostring(nResultOp);
end

local function DecomposeFunctions(aTranslation,sExpression)
	return sExpression:gsub("%s*(%a+%w*)%s*(%b())%s*",function(sName,sParams) return DecomposeFunction(aTranslation,sName,sParams:sub(2,-2)) end);
end

local function DecomposeParantheses(aTranslation,sExpression)
	return sExpression:gsub("%s*(%b())%s*",function(sNestedExpr) return DecomposeExpression(aTranslation,sNestedExpr:sub(2,-2)) end);
end

-- separate binary and unary operators
-- like "+" and "-" from "+-" (like from the source "1+-2") or
-- like "or" and "not" from "ornot" (like from the source "true or not false")
-- should contain one or no binary operators, but can contain multiple unary operators
-- bSearchBinary: if false then only do unary operators
local function ReadOperatorElements(sOperators, bSearchBinary)
	sTempOps = sOperators;
	local aOperators = {};
	local nFoundOpLen;
	local nTestLen;

	local nBinaryOpId = nil;
	if bSearchBinary then
		-- find the longest matching binary operator from the left
		nFoundOpLen = 0;
		for _, d in pairs(aOperationDefs) do
			if d[OPERATIONDEF_FORMAT] == OPERATION_FORMAT_BINARY then
				nTestLen = d[OPERATIONDEF_OPERATOR]:len();
				if (nTestLen > nFoundOpLen) and (sTempOps:sub(1,nTestLen) == d[OPERATIONDEF_OPERATOR]) then
					nBinaryOpId = d[OPERATIONDEF_ID];
					nFoundOpLen = nTestLen;
				end
			end
		end
		if nFoundOpLen > 0 then
			-- remove found binary operator from operators
			sTempOps = sTempOps:sub(nFoundOpLen+1);
			sTempOps = sTempOps:match("%s*(.*)"); -- remove leading whitespace
		end
	end
	aOperators[1] = nBinaryOpId;
	
	local aUnaryOpIds = nil;
	local nUnaryOpId;
	while sTempOps ~= nil and sTempOps ~= "" do
		-- find the longest matching unary operator from the left
		nUnaryOpId = nil;
		nFoundOpLen = 0;
		for _, d in pairs(aOperationDefs) do
			if d[OPERATIONDEF_FORMAT] == OPERATION_FORMAT_UNARY then
				nTestLen = d[OPERATIONDEF_OPERATOR]:len();
				if (nTestLen > nFoundOpLen) and (sTempOps:sub(1,nTestLen) == d[OPERATIONDEF_OPERATOR]) then
					nUnaryOpId = d[OPERATIONDEF_ID];
					nFoundOpLen = nTestLen;
				end
			end
		end
		if nFoundOpLen > 0 then
			if aUnaryOpIds == nil then
				aUnaryOpIds = {};
			end
			aUnaryOpIds[#aUnaryOpIds+1] = nUnaryOpId;
			-- remove found unary operator from operators
			sTempOps = sTempOps:sub(nFoundOpLen+1);
			sTempOps = sTempOps:match("%s*(.*)"); -- remove leading whitespace
		else
			-- didn't find any - just break off
			break;
		end
	end
	aOperators[2] = aUnaryOpIds;
	
	return aOperators;
end

-- create a table of Elements out of an expression
-- an operand Element wil contain the operand id value
-- an operators Element wil contain a table of 2 items:
-- [1]: binary operator id value
-- [2]: table of unary operator id values
local function ReadExpressionElements(sExpression)
	local sTempExpr = "!"..sExpression.."!"; -- add some character to start and end for pattern matching
	local sBinaryOpEndSearch = "^(.*)%[(%d+)%]%s*(.-)%s*%[(%d+)%](.-)$"; -- find match at the end
	local sUnaryOpStartSearch = "^(!)%s*(.-)%s*%[(%d+)%](.-)$"; -- find match at the start
	local aElements = {};
	local sPre, sFoundOperand1, sFoundOperator, sFoundOperand2, sPost;

	sPre, sFoundOperand1, sFoundOperator, sFoundOperand2, sPost = sTempExpr:match(sBinaryOpEndSearch);
	while sFoundOperator ~= nil do
		aElements[1] = tonumber(sFoundOperand1); -- 1 is always last operand1 index number found
		table.insert(aElements,2,tonumber(sFoundOperand2)); -- store second operand index number
		table.insert(aElements,2,ReadOperatorElements(sFoundOperator,true)); -- operator may be combination of binary and unary operators
		-- remove operand2 and operator(s) from search expression
		sTempExpr = sPre..OPNtostring(sFoundOperand1)..sPost;
		sPre, sFoundOperand1, sFoundOperator, sFoundOperand2, sPost = sTempExpr:match(sBinaryOpEndSearch);
	end

	-- we might have a unary operation left at the start like in the expression "NOT true or false"
	sPre, sFoundOperator, sFoundOperand1, sPost = sTempExpr:match(sUnaryOpStartSearch);
	if sFoundOperator ~= "" then
		table.insert(aElements,1,ReadOperatorElements(sFoundOperator,false)); -- operator must be a one or more unary operators
		if sPre == nil then
			sTempExpr = "";
		else
			sTempExpr = sPre;
		end
		if #aElements == 1 then
			-- no operand yet for this unary, so get it
			table.insert(aElements,2,tonumber(sFoundOperand1));
		else
			sTempExpr = sTempExpr..OPNtostring(sFoundOperand1);
		end
		if sPost ~= nil then
			sTempExpr = sTempExpr..sPost;
		end
	end

	return aElements;
end

local function FindBinaryOperationInPrioGroup(aElements,nElemIndex,aGroupOperatorDefIds)
	local aCombinedOperators = aElements[nElemIndex];
	if type(aCombinedOperators) == "table" then
		local nBinaryOperator = aCombinedOperators[1];
		local nGroupOperationDefId;
		for _, nGroupOperationDefId in pairs(aGroupOperatorDefIds) do
			if nGroupOperationDefId == nBinaryOperator then
				return nBinaryOperator;
			end
		end
	end
	return nil;
end

local function ProcessUnaryOperations(aTranslation,aElements,nElemIndex,aGroupOperatorDefIds,nScanDirection)
	local aCombinedOperators = aElements[nElemIndex];
	local bRemoveElement = false;
	if type(aCombinedOperators) == "table" then
		local aUnaryOperators = aCombinedOperators[2];
		if type(aUnaryOperators) == "table" then
			-- has unary operators
			local nCurrentUnary, nStartUnary, nEndUnary, nUnaryStep;
			local nUnaryOperator;
			local nGroupOperationDefId;
			local nOperationDef;
			local aParamOps;
			local sSource;
			local nResultOp;

			if nScanDirection == SCANDIRECTION_RIGHTLEFT then
				nStartUnary = #aUnaryOperators;
				nEndUnary = 1;
				nUnaryStep = -1;
			elseif nScanDirection == SCANDIRECTION_LEFTRIGHT then
				nStartUnary = 1;
				nEndUnary = #aUnaryOperators;
				nUnaryStep = 1;
			end

			nCurrentUnary = nStartUnary;
			repeat
				nUnaryOperator = aUnaryOperators[nCurrentUnary];
				for _, nGroupOperationDefId in pairs(aGroupOperatorDefIds) do
					if nGroupOperationDefId == nUnaryOperator then
						-- process found unary operator
						nOperationDef = GetOperationDefById(nGroupOperationDefId);

						aParamOps = {aElements[nElemIndex+1]};
						sSource = aOperationDefs[nOperationDef][OPERATIONDEF_OPERATOR]..OPNtostring(aParamOps[1]);

						-- create an operation for this unary operation (returns result operand number)
						nResultOp = CreateOperation(aTranslation,sSource,nOperationDef,aParamOps);

						-- replace source operand with result operand
						aElements[nElemIndex+1] = nResultOp;
					
						-- remove processed unary operator from unary operators
						table.remove(aUnaryOperators,nCurrentUnary);
						if nScanDirection == SCANDIRECTION_LEFTRIGHT then
							nCurrentUnary = nCurrentUnary-1;
							nEndUnary = nEndUnary-1;
						end
						
						break;
					end
				end
				nCurrentUnary = nCurrentUnary+nUnaryStep;
			until (nScanDirection == SCANDIRECTION_LEFTRIGHT and nCurrentUnary > nEndUnary) or (nScanDirection == SCANDIRECTION_RIGHTLEFT and nCurrentUnary < nEndUnary)
			
			if #aUnaryOperators == 0 then
				aCombinedOperators[2] = nil;
				bRemoveElement = (aCombinedOperators[1] == nil);
			else
				aCombinedOperators[2] = aUnaryOperators;
			end
			aElements[nElemIndex] = aCombinedOperators;
		end
	end
	return bRemoveElement;
end

local function ProcessBinaryOperation(aTranslation,aElements,nElemIndex,aGroupOperatorDefIds)
	local aCombinedOperators = aElements[nElemIndex];
	local bRemoveElement = false;
	if type(aCombinedOperators) == "table" then
		local nBinaryOperator = aCombinedOperators[1];
		if type(nBinaryOperator) == "number" then
			-- has binary operator
			local nOperationDef;
			local aParamOps;
			local sSource;
			local nResultOp;
			
			nOperationDef = GetOperationDefById(nBinaryOperator);
			
			aParamOps = {aElements[nElemIndex-1],aElements[nElemIndex+1]};
			sSource = OPNtostring(aParamOps[1])..aOperationDefs[nOperationDef][OPERATIONDEF_OPERATOR]..OPNtostring(aParamOps[2]);

			-- create an operation for this binary operation (returns result operand number)
			nResultOp = CreateOperation(aTranslation,sSource,nOperationDef,aParamOps);
			
			aCombinedOperators[1] = nil;
			aElements[nElemIndex-1] = nResultOp;
			aElements[nElemIndex] = aCombinedOperators;
			aElements[nElemIndex+1] = nResultOp;
			
			bRemoveElement = true;
		end
	end
	return bRemoveElement;
end

local function CreateExpressionFromElements(aElements)
	local sResult = "";
	local nOperationDef;
	for _, e in pairs(aElements) do
		if type(e) == "number" then
			-- operand number
			sResult = sResult..OPNtostring(e);
		elseif type(e) == "table" then
			if e[1] ~= nil then
				-- binary operatordefid
				nOperationDef = GetOperationDefById(e[1]);
				sResult = sResult..aOperationDefs[nOperationDef][OPERATIONDEF_OPERATOR];
			end
			if e[2] ~= nil then
				-- unary operatordefid(s)
				for _, o in pairs(e[2]) do
					nOperationDef = GetOperationDefById(o);
					sResult = sResult..aOperationDefs[nOperationDef][OPERATIONDEF_OPERATOR];
				end
			end
		end
	end
	return sResult;
end

local function CreateOperations(aTranslation,sExpression)
	local aElements = ReadExpressionElements(sExpression);
	local aOperationPriorityGroup;
	local aOperationDefIds;
	local nScanDirection;
	local nScanFormat;
	local nCurrentElem, nStartElem, nEndElem, nElemStep;
	local nBinaryOperatorId;

	for _, aOperationPriorityGroup in pairs(aOperationPriorityGroups) do
		aOperationDefIds = aOperationPriorityGroup[OPPRIOGRP_OPERATIONDEF_IDS]; -- need to search for these IDs
		nScanDirection = aOperationPriorityGroup[OPPRIOGRP_SCANDIRECTION]; -- from which direction to search
		nScanFormat = aOperationPriorityGroup[OPPRIOGRP_SCANFORMAT]; -- search binary or unary operations
		
		if nScanDirection == SCANDIRECTION_RIGHTLEFT then
			nStartElem = #aElements-1; -- last Element should be operand, so just before are first operators
			nEndElem = 1; -- last operator Element should be either 1 (only unary) or 2 (ending with binary/unary)
			nElemStep = -2; -- skip an operand and an operator(s) Element to the left
		elseif nScanDirection == SCANDIRECTION_LEFTRIGHT then
			if type(aElements[1]) == "number" then
				-- Element 1 operand, 2 operator(s), etc
				nStartElem = 2;
			else
				-- Element 1 unary operator, 2 operand, 3 operator(s) etc
				if nScanFormat == OPERATION_FORMAT_BINARY then
					nStartElem = 3; -- skip unary operator at the start
				elseif nScanFormat == OPERATION_FORMAT_UNARY then
					nStartElem = 1; -- start with the unary operator at 1
				end
			end
			nEndElem = #aElements-1; -- last Element should be an operand, so just before
			nElemStep = 2; -- skip an operand and an operator(s) Element to the right
		end

		nCurrentElem = nStartElem;
		repeat
			if nScanFormat == OPERATION_FORMAT_BINARY then
				nBinaryOperatorId = FindBinaryOperationInPrioGroup(aElements,nCurrentElem,aOperationDefIds);
				if nBinaryOperatorId ~= nil then
					-- found a binary operation which is in current priority group
					
					-- right to left ^: -[2]^-[3]^+[4] -> -[2]^-([3]^+[4]) -> (-([2]^(-([3]^(+[4])))))
					-- left to right *: -[2]*-[3]*+[4] -> ((-[2])*(-[3]))*+[4] -> (((-[2])*(-[3]))*(+[4]))

					-- always process all unaries (if still any) before processing the binary operation
					ProcessUnaryOperations(aTranslation,aElements,nCurrentElem,aOperationPriorityGroups[2][OPPRIOGRP_OPERATIONDEF_IDS],SCANDIRECTION_RIGHTLEFT); --group 2 contains the unary operations

					if ProcessBinaryOperation(aTranslation,aElements,nCurrentElem,nBinaryOperatorId) then
						-- should remove operation Element if here
						-- this means that binary has been processed and should be removed
						table.remove(aElements,nCurrentElem); -- remove operation
						table.remove(aElements,nCurrentElem); -- remove operand2
						if nScanDirection == SCANDIRECTION_LEFTRIGHT then
							nCurrentElem = nCurrentElem-2;
							nEndElem = nEndElem-2;
						end
					end
				end
			elseif nScanFormat == OPERATION_FORMAT_UNARY then
				if ProcessUnaryOperations(aTranslation,aElements,nCurrentElem,aOperationDefIds,nScanDirection) then
					-- should remove operation Element if here
					-- this means that the unary at the start has been processed and can now be removed
					table.remove(aElements,nCurrentElem); -- remove operation
					if nScanDirection == SCANDIRECTION_LEFTRIGHT then
						nCurrentElem = nCurrentElem-1;
						nEndElem = nEndElem-1;
					end
				end
			end
			nCurrentElem = nCurrentElem+nElemStep;
		until (nScanDirection == SCANDIRECTION_LEFTRIGHT and nCurrentElem > nEndElem) or (nScanDirection == SCANDIRECTION_RIGHTLEFT and nCurrentElem < nEndElem)
	end

	return CreateExpressionFromElements(aElements);
end

-- recursive function which decomposes (parts of) an expression
function DecomposeExpression(aTranslation,sExpression)

	if aTranslation[TRANS_ERRCODE] ~= 0 then
		-- error state: don't continue
		return sExpression;
	end

	local nResultOp;
	local sTempExpr = sExpression;
	
	while true do
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			break;
		end
	
		-- contains only [n] ?
		nResultOp = OPStonumber(sTempExpr);
		if nResultOp ~= nil then
			-- expression already contains final result (in operand)
			sTempExpr = OPNtostring(nResultOp);
			break;
		end
	
		-- process functions funcname(expr1,expr2,expr3,..)
		sTempExpr = DecomposeFunctions(aTranslation,sTempExpr);
	
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			-- error state: don't continue
			break;
		end
		-- contains only [n] ?
		nResultOp = OPStonumber(sTempExpr);
		if nResultOp ~= nil then
			-- result is from a single function
			sTempExpr = OPNtostring(nResultOp);
			break;
		end
	
		-- process parantheses (expr)
		sTempExpr = DecomposeParantheses(aTranslation,sTempExpr);
	
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			-- error state: don't continue
			break;
		end
		-- contains only [n] ?
		nResultOp = OPStonumber(sTempExpr);
		if nResultOp ~= nil then
			-- result is from a single nested ()
			sTempExpr = OPNtostring(nResultOp);
			break;
		end
	
		-- create operands from remaining constant types (number, boolean)
		-- not sooner because of possible numbers in function names
		sTempExpr = CreateConstantOperands(aTranslation,sTempExpr);
	
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			-- error state: don't continue
			break;
		end
		-- contains only [n] ?
		nResultOp = OPStonumber(sTempExpr);
		if nResultOp ~= nil then
			--- result is from a single constant
			sTempExpr = OPNtostring(nResultOp);
			break;
		end
	
		sTempExpr = CreateOperations(aTranslation,sTempExpr);

		if aTranslation[TRANS_ERRCODE] ~= 0 then
			-- error state: don't continue
			break;
		end
		-- contains only [n] ?
		nResultOp = OPStonumber(sTempExpr);
		if nResultOp ~= nil then
			-- result is from simple operations
			sTempExpr = OPNtostring(nResultOp);
			break;
		end

		-- here when not finalized completely because of a syntax error.
		aTranslation[TRANS_ERRCODE] = ERR_EXPRESSIONSYNTAX;
		break;
	end

	return sTempExpr;
end

-- translates an expression to an intermediate object, given a definition of runtime variables.
-- resulting object may be evaluated/executed (see ExecuteExpression()), while providing runtime variables with values.
function TranslateExpression(sExpression,aVariables)

	-- basic initialization of translation object
	local aTranslation = {};
	aTranslation[TRANS_EXPRESSION] = sExpression; -- Original expression which translation will represent
	aTranslation[TRANS_EXPRWIP] = sExpression; -- expression work in progress: expression in various states of translation
	aTranslation[TRANS_ERRCODE] = 0; -- indicating errors during translation or execution
	aTranslation[TRANS_OPERATIONS] = {}; -- Operations table
	aTranslation[TRANS_OPERANDS] = {}; -- Operands table
	aTranslation[TRANS_VARIABLES] = {}; -- Variables table

	-- populate variable table with supplied variables
	local aTransVars = aTranslation[TRANS_VARIABLES];
	for i, v in pairs(aVariables) do
		aTransVars[i] = {};
		aTransVars[i][VAR_NAME] = "$"..v[VAR_NAME]:upper();
		aTransVars[i][VAR_VALUE] = v[VAR_VALUE]; -- initial value
		aTransVars[i][VAR_OPERANDS] = {}; -- operands registration table
	end

	-- create operands for string constants
	CreateStringConstants(aTranslation);

	-- convert expression to uppercase now we have processed the strings
	aTranslation[TRANS_EXPRWIP] = aTranslation[TRANS_EXPRWIP]:upper();
	
	-- replace CalcStat shorts: @StatName etc with function call expressions
	ReplaceCalcStatShorts(aTranslation);
	
	-- create operands for variables
	CreateVariableOperands(aTranslation);

	-- extract operations - returns "[n]", with indicating operand index n containing the final result
	aTranslation[TRANS_EXPRWIP] = DecomposeExpression(aTranslation,aTranslation[TRANS_EXPRWIP]);
		
	return aTranslation;
end

local function GetOperandValue(aTranslation,nOp)
	local aOperand = aTranslation[TRANS_OPERANDS][nOp];
	
	if aOperand[OPERAND_VALUE] == nil then
		-- need to get the value from an operation (a constant would always have a value)
		if aOperand[OPERAND_OPERATION] == OPERAND_OPERATION_GETVAR then
			-- need to get value from variable (DependsOn should contain only 1 variable reference)
			local aVar, nCheckBits, nBit;
			nCheckBits = aOperand[OPERAND_DEPENDSON];
			for _, aVar in pairs(aTranslation[TRANS_VARIABLES]) do
				nBit = nCheckBits % 2;
				if nBit > 0 then
					-- operand depends on this variable
					aOperand[OPERAND_VALUE] = aVar[VAR_VALUE];
					break;
				end
				nCheckBits = (nCheckBits-nBit) / 2;
			end
		else
			-- we should get the value when we execute its operation
			ExecuteOperation(aTranslation,aTranslation[TRANS_OPERATIONS][aOperand[OPERAND_OPERATION]]);
		end
	end
	
	return aOperand[OPERAND_VALUE];
end

local adjustfordisplay;
local stringformatvalue;
-- executes a single operation and stores the outcome in a (intermediate) result operand
function ExecuteOperation(aTranslation,aOperation)
	local aOperationDef = aOperationDefs[aOperation[OPERATION_DEFINITION]];
	local nOpDefId = aOperationDef[OPERATIONDEF_ID];
	local nOpDefFormat = aOperationDef[OPERATIONDEF_FORMAT];
	local aOperands = aOperation[OPERATION_OPERANDS];
	local nOperandCount = #aOperands;
	local xOperand1, xOperand2, xOperand3;
	local xResult;
	
	if aTranslation[TRANS_ERRCODE] ~= 0 then
		return;
	end
	
	if nOpDefFormat == OPERATION_FORMAT_UNARY then
		if nOperandCount ~= 1 then
			aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
			return;
		end
		xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			return;
		end
		if nOpDefId == UNARYOP_LOGICNEG then
			-- logic negate
			if type(xOperand1) == "boolean" or type(xOperand1) == "number" then 
				xResult = (not xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == UNARYOP_NEGATE then
			-- number negate
			if type(xOperand1) == "number" then 
				xResult = -xOperand1;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == UNARYOP_POSITIVE then
			-- number positive
			if type(xOperand1) == "number" then 
				xResult = xOperand1; -- just copy
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		end
	elseif nOpDefFormat == OPERATION_FORMAT_BINARY then
		if nOperandCount ~= 2 then
			aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
			return;
		end
		xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			return;
		end
		xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
		if aTranslation[TRANS_ERRCODE] ~= 0 then
			return;
		end
		if nOpDefId == BINARYOP_CONCAT then
			-- concatenate 2 operands into a string (first need to be a string)
			if (type(xOperand1) == "boolean" or type(xOperand1) == "number" or type(xOperand1) == "string") and (type(xOperand2) == "boolean" or type(xOperand2) == "number" or type(xOperand2) == "string") then 
				if type(xOperand1) == "boolean" then
					if xOperand1 then
						xOperand1 = "True";
					else
						xOperand1 = "False";
					end
				end
				if type(xOperand2) == "boolean" then
					if xOperand2 then
						xOperand2 = "True";
					else
						xOperand2 = "False";
					end
				end
				xResult = xOperand1..""..xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_OR then
			if (type(xOperand1) == "boolean" or type(xOperand1) == "number") and (type(xOperand2) == "boolean" or type(xOperand2) == "number") then 
				xResult = (xOperand1 or xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_AND then
			if (type(xOperand1) == "boolean" or type(xOperand1) == "number") and (type(xOperand2) == "boolean" or type(xOperand2) == "number") then 
				xResult = (xOperand1 and xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_GREQTH then
			if (type(xOperand1) == "number" and type(xOperand2) == "number") or (type(xOperand1) == "string" and type(xOperand2) == "string") then 
				xResult = (xOperand1 >= xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_SMEQTH then
			if (type(xOperand1) == "number" and type(xOperand2) == "number") or (type(xOperand1) == "string" and type(xOperand2) == "string") then 
				xResult = (xOperand1 <= xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_EQUALS then
			xResult = (xOperand1 == xOperand2);
		elseif nOpDefId == BINARYOP_NOTEQUALS then
			xResult = (xOperand1 ~= xOperand2);
		elseif nOpDefId == BINARYOP_GRTH then
			if (type(xOperand1) == "number" and type(xOperand2) == "number") or (type(xOperand1) == "string" and type(xOperand2) == "string") then 
				xResult = (xOperand1 > xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_SMTH then
			if (type(xOperand1) == "number" and type(xOperand2) == "number") or (type(xOperand1) == "string" and type(xOperand2) == "string") then 
				xResult = (xOperand1 < xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_SUBSTRACT then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 - xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_ADD then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 + xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_MOD then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 % xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_DIV then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = math.floor(xOperand1 / xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_MULTIPLY then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 * xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_DIVIDE then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 / xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == BINARYOP_POWER then
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = xOperand1 ^ xOperand2;
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		end
	elseif nOpDefFormat == OPERATION_FORMAT_FUNCTION then
		if nOpDefId == FUNCTIONOP_IIF then
			if nOperandCount ~= 3 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if (type(xOperand1) == "boolean" or type(xOperand1) == "number") then 
				if xOperand1 then
					xResult = GetOperandValue(aTranslation,aOperands[2]);
				else
					xResult = GetOperandValue(aTranslation,aOperands[3]);
				end
				if aTranslation[TRANS_ERRCODE] ~= 0 then
					return;
				end
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_CALCSTAT then
			if nOperandCount < 1 or nOperandCount > 3 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			xOperand2 = nil;
			xOperand3 = nil;
			if nOperandCount >= 2 then
				xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
				if aTranslation[TRANS_ERRCODE] ~= 0 then
					return;
				end
				if nOperandCount == 3 then
					xOperand3 = GetOperandValue(aTranslation,aOperands[3]);
					if aTranslation[TRANS_ERRCODE] ~= 0 then
						return;
					end
				end
			end
			if (type(xOperand1) == "string") and (type(xOperand2) == "nil" or type(xOperand2) == "number") then 
				if xOperand3 == nil then
					xResult = CalcStat(xOperand1,xOperand2);
				else
					xResult = CalcStat(xOperand1,xOperand2,xOperand3);
				end
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_ABS then
			if nOperandCount ~= 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then 
				xResult = math.abs(xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_CEIL then
			if nOperandCount ~= 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then 
				xResult = math.ceil(xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_FLOOR then
			if nOperandCount ~= 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then 
				xResult = math.floor(xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_MAX then
			if nOperandCount ~= 2 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = math.max(xOperand1,xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_MIN then
			if nOperandCount ~= 2 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" and type(xOperand2) == "number" then 
				xResult = math.min(xOperand1,xOperand2);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_LOG10 then
			if nOperandCount ~= 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then 
				xResult = math.log10(xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_SQRT then
			if nOperandCount ~= 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then 
				xResult = math.sqrt(xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_AFD then
			if nOperandCount < 1 or nOperandCount > 2 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if nOperandCount == 2 then
				xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
				if aTranslation[TRANS_ERRCODE] ~= 0 then
					return;
				end
			else
				xOperand2 = "CORR"; --default: 0.0002 correction
			end
			if type(xOperand2) == "string" then 
				xResult = adjustfordisplay(xOperand2,xOperand1,nil);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_FORMAT then
			if nOperandCount < 1 or nOperandCount > 2 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if nOperandCount == 2 then
				xOperand2 = GetOperandValue(aTranslation,aOperands[2]);
				if aTranslation[TRANS_ERRCODE] ~= 0 then
					return;
				end
			elseif type(xOperand1) == "number" then
				xOperand2 = "%.0f"; --default: number without decimals
			elseif type(xOperand1) == "string" then
				xOperand2 = "%s"; --default: just same string
			end
			if (type(xOperand1) == "number" or type(xOperand1) == "string") and (type(xOperand2) == "string") then 
				xResult = stringformatvalue(xOperand2,xOperand1);
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_SWITCH then
			local nOp;
			for nOp = 1, nOperandCount, 2 do
				-- process logical + result operand pairs
				-- default logical is a constant true, so will always be the result once encountered
				if nOp+1 > nOperandCount then
					aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
					return;
				end
				xOperand1 = GetOperandValue(aTranslation,aOperands[nOp]);
				if aTranslation[TRANS_ERRCODE] ~= 0 then
					return;
				end
				if type(xOperand1) == "boolean" then
					if xOperand1 then
						xOperand2 = GetOperandValue(aTranslation,aOperands[nOp+1]);
						if aTranslation[TRANS_ERRCODE] ~= 0 then
							return;
						end
						xResult = xOperand2;
						break;
					end
				else
					aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
					break;
				end
			end
		elseif nOpDefId == FUNCTIONOP_SESEL then
			if nOperandCount < 2 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			xOperand1 = GetOperandValue(aTranslation,aOperands[1]);
			if aTranslation[TRANS_ERRCODE] ~= 0 then
				return;
			end
			if type(xOperand1) == "number" then
				 -- compare number value minus a little for small discrepancies
				 -- like 79.0000001 should be treated like 79 so <= 79
				xOperand1 = xOperand1-DblCalcDev;
			end
			if (type(xOperand1) == "number" or type(xOperand1) == "string") then 
				local nOp;
				for nOp = 2, nOperandCount, 2 do
					xOperand2 = GetOperandValue(aTranslation,aOperands[nOp]);
					if aTranslation[TRANS_ERRCODE] ~= 0 then
						return;
					end
					if nOp == nOperandCount then
						-- last operand is default value
						xResult = xOperand2;
					else
						if (type(xOperand1) == "number" and type(xOperand2) == "number") or (type(xOperand1) == "string" and type(xOperand2) == "string") then
							if xOperand1 <= xOperand2 then
								xOperand3 = GetOperandValue(aTranslation,aOperands[nOp+1]);
								if aTranslation[TRANS_ERRCODE] ~= 0 then
									return;
								end
								xResult = xOperand3;
								break;
							end
						else
							aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
							break;
						end
					end
				end
			else
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPTYPE;
			end
		elseif nOpDefId == FUNCTIONOP_WITH then
			if nOperandCount < 1 then
				aTranslation[TRANS_ERRCODE] = ERR_INVALIDOPCOUNT;
				return;
			end
			local nOp;
			for nOp = 1, nOperandCount, 2 do
				if nOp == nOperandCount then
					-- last operand is result value
					xOperand1 = GetOperandValue(aTranslation,aOperands[nOp]);
					if aTranslation[TRANS_ERRCODE] ~= 0 then
						return;
					end
					xResult = xOperand1;
				else
					-- process variable assignments
					xOperand2 = GetOperandValue(aTranslation,aOperands[nOp+1]);
					if aTranslation[TRANS_ERRCODE] ~= 0 then
						return;
					end
					SetExpressionVariable(aTranslation,aTranslation[TRANS_OPERANDS][aOperands[nOp]][OPERAND_SOURCE],xOperand2);
				end
			end
		end
	end
	
	-- store result in result operand
	aTranslation[TRANS_OPERANDS][aOperation[OPERATION_RESULTOP]][OPERAND_VALUE] = xResult;
end

-- execute translated expression and return result
function ExecuteExpression(aTranslation)
	if aTranslation[TRANS_ERRCODE] ~= 0 then
		return "ErrCode:"..aTranslation[TRANS_ERRCODE];
	end
	-- return result from last created operand
	local xResultValue = GetOperandValue(aTranslation,#aTranslation[TRANS_OPERANDS]);
	if aTranslation[TRANS_ERRCODE] ~= 0 then
		return "ErrCode:"..aTranslation[TRANS_ERRCODE];
	end
	return xResultValue;
end

-- sets a variable to a value and nullifies depending operands' values, so they'll be recalculated at execution time
function SetExpressionVariable(aTranslation,sVarName,xNewValue)
	local sSearchVar;
	if sVarName:sub(1,1) == "$" then
		sSearchVar = sVarName:upper();
	else
		sSearchVar = "$"..sVarName:upper();
	end
	local aVar, xOldValue, bNew;
	for _, aVar in pairs(aTranslation[TRANS_VARIABLES]) do
		if aVar[VAR_NAME] == sSearchVar then
			xOldValue = aVar[VAR_VALUE];
			if type(xOldValue) ~= type(xNewValue) then
				bNew = true;
			else
				bNew = (xOldValue ~= xNewValue);
			end
			if bNew then
				-- store new value
				aVar[VAR_VALUE] = xNewValue;
				-- nullify registered operands values
				local aTransOperands = aTranslation[TRANS_OPERANDS];
				local nOp;
				for _, nOp in pairs(aVar[VAR_OPERANDS]) do
					-- sets value of registered operand to nil
					aTransOperands[nOp][OPERAND_VALUE] = nil;
				end
				return xOldValue;
			end
			return nil;
		end
	end
	return nil;
end

function GetExpressionError(aTranslation)
	if aTranslation[TRANS_ERRCODE] == 0 then
		return 0, nil, nil;
	else
		return aTranslation[TRANS_ERRCODE], aErrMsgs[aTranslation[TRANS_ERRCODE]], aTranslation[TRANS_EXPRWIP];
	end
end

-- to be used by other modules
p.TranslateExpression = TranslateExpression;
p.ExecuteExpression = ExecuteExpression;
p.SetExpressionVariable = SetExpressionVariable;
p.GetExpressionError = GetExpressionError;

-- ************************ End CalcStat Expressions **************************

-- ******************** Start Formatting/Display Support **********************

-- small number correction for display purposes only (like in-game)
local function correctvalue(nValue)

	if nValue < -DblCalcDev then
		return nValue-0.0002
	elseif nValue > DblCalcDev then
		return nValue+0.0002
	else
		return 0
	end

end

-- uses template Worth to output value(format ..ggggssscc) as currency (..gggg gold sss silver cc copper)
local function currencyvalue(nValue,frame)

	local nTemp = math.floor(nValue+0.5+DblCalcDev);
	local nCopper = nTemp%100;
	nTemp = (nTemp-nCopper)/100;
	local nSilver = nTemp%1000;
	nTemp = (nTemp-nSilver)/1000;
	local nGold = nTemp;
	
	local aTemplateArgs;
	if nGold > 0 then
		aTemplateArgs = {g=nGold,s=nSilver,c=nCopper};
	elseif nSilver > 0 then
		aTemplateArgs = {s=nSilver,c=nCopper};
	else
		aTemplateArgs = {c=nCopper};
	end

	return frame:expandTemplate{title="Worth",args=aTemplateArgs};

end

function adjustfordisplay(sOption,xValue,frame)

	local xTemp = xValue;

	if type(xTemp) == "number" then
		-- optional display options for percentages
		-- for display purposes only (same as in-game)
		if sOption == "MULTIP" then
			-- multiplier percentage
			xTemp = correctvalue(xTemp-100);
		elseif sOption == "MULTIP100" then
			-- multiplier percentage & multiply by 100
			xTemp = correctvalue(xTemp*100-100);
		elseif sOption == "ADDP" then
			-- additive percentage
			xTemp = correctvalue(xTemp);
		elseif sOption == "ADDP100" then
			-- additive percentage & multiply by 100
			xTemp = correctvalue(xTemp*100);
		elseif sOption == "WORTH" then
			-- currency display
			if frame ~= nil then
				xTemp = currencyvalue(xTemp,frame);
			end
		elseif sOption == "CORR" then
			-- fractured value which needs a correction for rounded display
			xTemp = correctvalue(xTemp);
		elseif sOption == "NOADJ" then
			-- no adjustment
		end
	end

	return xTemp;

end

local function AddCommas(pre,post) return pre..post:reverse():gsub("(%d%d%d)","%1"..","):reverse(); end
local function SplitDecimal(str) return str:gsub("^(%d)(%d+)",AddCommas); end
local function ReformatThousSep(str) return str:gsub("[%d%.]+",SplitDecimal); end

-- adds trailing zero removal option (set minimum precision): %.0.3f
-- adds thousand separator option to formatting: %'d
function stringformatvalue(sFormat,xValue)

	if type(xValue) == "number" then
		local sPreFormat = sFormat;

		local bPostThousSep = string.find(sPreFormat,"%%'");
		if bPostThousSep then
			sPreFormat = string.gsub(sPreFormat,"%%'","%%");
		end

		local bPostFormatTrailZeroes = string.find(sPreFormat,".%d+.%d+%a");
		local nTrailZeroesMinPrec;

		if bPostFormatTrailZeroes then
			local _, _, sTrailZeroesMinPrec = string.find(sPreFormat,".(%d+).%d+%a");
			nTrailZeroesMinPrec = tonumber(sTrailZeroesMinPrec);
			sPreFormat = string.gsub(sPreFormat,".%d+(.%d+%a)","%1");
		end

		local sPostFormat = string.format(sPreFormat,xValue);

		if bPostFormatTrailZeroes then
			local sTrailZeroesGetDecimals = "%d+(.%d+)";
			local sTrailZeroesSetDecimals = "(%d+).%d+";
			local _, _, sDotDecimals = string.find(sPostFormat,sTrailZeroesGetDecimals);

			while sDotDecimals ~= nil and string.len(sDotDecimals) > nTrailZeroesMinPrec+1 and string.sub(sDotDecimals,-1,-1) == "0" do
				sDotDecimals = string.sub(sDotDecimals,1,-2);
			end

			if sDotDecimals == "." then
				sPostFormat = string.gsub(sPostFormat,sTrailZeroesSetDecimals,"%1");
			elseif sDotDecimals ~= nil then
				sPostFormat = string.gsub(sPostFormat,sTrailZeroesSetDecimals,"%1"..sDotDecimals);
			end
		end

		if bPostThousSep then
			sPostFormat = ReformatThousSep(sPostFormat);
		end

		return sPostFormat;
	else
		return string.format(sFormat,xValue);
	end
	
end

-- ********************* End Formatting/Display Support ***********************
