import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

import "Giseldah.CalcStat.CSExpres";

local function ShowHelpOutput()
	ltbOutput:SetText(
						"Field $L(evel): enter a number which will be assigned to variable $L.\n"..
						"Field to: if you enter a number here then the expression will be run multiple times with a changing $L.\n"..
						"Field $N or $C: if you enter a number then this will be assigned to $N and otherwise it will be $C.\n"..
						"\n"..
						"See https://lotro-wiki.com/index.php/Module:CalcStat/doc#Expressions for some more information.\n"..
						"\n"..
						"List of all available Player variables:\n"..
						"-- Character --\n"..
						"$PlayerLevel\n"..
						"$PlayerName\n"..
						"$PlayerRace, $PlayerRaceName\n"..
						"$PlayerClass, $PlayerClassName\n"..
						"$PlayerMorale, $PlayerPower (current)\n"..
						"\n"..
						"-- Basic Stats --\n"..
						"$PlayerMaxMorale, $PlayerICMR, $PlayerNCMR\n"..
						"$PlayerMaxPower, $PlayerICPR, $PlayerNCPR\n"..
						"$PlayerArmour\n"..
						"$PlayerMight\n"..
						"$PlayerAgility\n"..
						"$PlayerVitality\n"..
						"$PlayerWill\n"..
						"$PlayerFate\n"..
						"-- Offence --\n"..
						"$PlayerCritHit, $PlayerMelCritHit, $PlayerRngCritHit, $PlayerTacCritHit\n"..
						"$PlayerFinesse\n"..
						"$PlayerPhyMas, $PlayerMelOff, $PlayerRngOff\n"..
						"$PlayerTacMas, $PlayerTacOff, $PlayerOutHeal\n"..
						"-- Defence --\n"..
						"$PlayerResist, $PlayerDiseaseResist, $PlayerFearResist, $PlayerPoisonResist, $PlayerWoundResist\n"..
						"$PlayerCritDef, $PlayerMelCritDef, $PlayerRngCritDef, $PlayerTacCritDef\n"..
						"$PlayerInHeal\n"..
						"  -- Avoidance --\n"..
						"  $PlayerBlock\n"..
						"  $PlayerParry\n"..
						"  $PlayerEvade\n"..
						"  -- Mitigations --\n"..
						"    -- Damage Source --\n"..
						"    $PlayerMelDef\n"..
						"    $PlayerRngDef\n"..
						"    $PlayerTacDef\n"..
						"    -- Damage Type --\n"..
						"    $PlayerPhyMit, $PlayerOFMit\n"..
						"    $PlayerTacMit, $PlayerFireMit, $PlayerLightningMit, $PlayerFrostMit, $PlayerAcidMit, $PlayerShadowMit"
					);
end

local function ClearMyExpression()
	ltbExpression:SetText("");
	ltbOutput:SetText("");
end

local function SetPlayerVariables(aTranslation)

	-- interface to CalcStat classnames
	CSClassNames = {
		[214] = "Beorning",
		[215] = "Brawler",
		[40] = "Burglar",
		[24] = "Captain",
		[172] = "Champion",
		[23] = "Guardian",
		[162] = "Hunter",
		[185] = "LoreMaster",
		[31] = "Minstrel",
		[193] = "RuneKeeper",
		[194] = "Warden"
	};

	-- interface to CalcStat racenames
	CSRaceNames = {
		[114] = "Beorning",
		[73] = "Dwarf",
		[65] = "Elf",
		[117] = "HighElf",
		[81] = "Hobbit",
		[23] = "Man",
		[120] = "StoutAxe"
	};

	-- Character
	SetExpressionVariable(aTranslation,"PlayerClass",Player:GetClass());
	SetExpressionVariable(aTranslation,"PlayerClassName",CSClassNames[Player:GetClass()]);
	SetExpressionVariable(aTranslation,"PlayerLevel",Player:GetLevel());
	SetExpressionVariable(aTranslation,"PlayerMaxMorale",Player:GetMaxMorale());
	SetExpressionVariable(aTranslation,"PlayerMaxPower",Player:GetMaxPower());
	SetExpressionVariable(aTranslation,"PlayerMorale",Player:GetMorale());
	SetExpressionVariable(aTranslation,"PlayerName",Player:GetName());
	SetExpressionVariable(aTranslation,"PlayerPower",Player:GetPower());
	SetExpressionVariable(aTranslation,"PlayerRace",Player:GetRace());
	SetExpressionVariable(aTranslation,"PlayerRaceName",CSRaceNames[Player:GetRace()]);
	
	-- Basic Stats
	SetExpressionVariable(aTranslation,"PlayerICMR",PlayerAttrib:GetInCombatMoraleRegeneration()*60);
	SetExpressionVariable(aTranslation,"PlayerNCMR",PlayerAttrib:GetOutOfCombatMoraleRegeneration()*60);
	SetExpressionVariable(aTranslation,"PlayerICPR",PlayerAttrib:GetInCombatPowerRegeneration()*60);
	SetExpressionVariable(aTranslation,"PlayerNCPR",PlayerAttrib:GetOutOfCombatPowerRegeneration()*60);
	SetExpressionVariable(aTranslation,"PlayerArmour",PlayerAttrib:GetArmor());
	SetExpressionVariable(aTranslation,"PlayerMight",PlayerAttrib:GetMight());
	SetExpressionVariable(aTranslation,"PlayerAgility",PlayerAttrib:GetAgility());
	SetExpressionVariable(aTranslation,"PlayerVitality",PlayerAttrib:GetVitality());
	SetExpressionVariable(aTranslation,"PlayerWill",PlayerAttrib:GetWill());
	SetExpressionVariable(aTranslation,"PlayerFate",PlayerAttrib:GetFate());

	-- Offence
	SetExpressionVariable(aTranslation,"PlayerCritHit",PlayerAttrib:GetBaseCriticalHitChance());
	SetExpressionVariable(aTranslation,"PlayerMelCritHit",PlayerAttrib:GetMeleeCriticalHitChance());
	SetExpressionVariable(aTranslation,"PlayerRngCritHit",PlayerAttrib:GetRangeCriticalHitChance());
	SetExpressionVariable(aTranslation,"PlayerTacCritHit",PlayerAttrib:GetTacticalCriticalHitChance());
	SetExpressionVariable(aTranslation,"PlayerFinesse",PlayerAttrib:GetFinesse());
	SetExpressionVariable(aTranslation,"PlayerPhyMas",math.max(PlayerAttrib:GetMeleeDamage(),PlayerAttrib:GetRangeDamage()));
	SetExpressionVariable(aTranslation,"PlayerMelOff",PlayerAttrib:GetMeleeDamage());
	SetExpressionVariable(aTranslation,"PlayerRngOff",PlayerAttrib:GetRangeDamage());
	SetExpressionVariable(aTranslation,"PlayerTacMas",PlayerAttrib:GetTacticalDamage());
	SetExpressionVariable(aTranslation,"PlayerTacOff",PlayerAttrib:GetTacticalDamage());
	SetExpressionVariable(aTranslation,"PlayerOutHeal",PlayerAttrib:GetOutgoingHealing());

	-- Defence
	SetExpressionVariable(aTranslation,"PlayerResist",PlayerAttrib:GetBaseResistance());
	SetExpressionVariable(aTranslation,"PlayerDiseaseResist",PlayerAttrib:GetDiseaseResistance());
	SetExpressionVariable(aTranslation,"PlayerFearResist",PlayerAttrib:GetFearResistance());
	SetExpressionVariable(aTranslation,"PlayerPoisonResist",PlayerAttrib:GetPoisonResistance());
	SetExpressionVariable(aTranslation,"PlayerWoundResist",PlayerAttrib:GetWoundResistance());
	SetExpressionVariable(aTranslation,"PlayerCritDef",PlayerAttrib:GetBaseCriticalHitAvoidance());
	SetExpressionVariable(aTranslation,"PlayerMelCritDef",PlayerAttrib:GetMeleeCriticalHitAvoidance());
	SetExpressionVariable(aTranslation,"PlayerRngCritDef",PlayerAttrib:GetRangeCriticalHitAvoidance());
	SetExpressionVariable(aTranslation,"PlayerTacCritDef",PlayerAttrib:GetTacticalCriticalHitAvoidance());
	SetExpressionVariable(aTranslation,"PlayerInHeal",PlayerAttrib:GetIncomingHealing());
	-- Avoidance
	SetExpressionVariable(aTranslation,"PlayerBlock",PlayerAttrib:GetBlock());
	SetExpressionVariable(aTranslation,"PlayerParry",PlayerAttrib:GetParry());
	SetExpressionVariable(aTranslation,"PlayerEvade",PlayerAttrib:GetEvade());
	-- Mitigations
	-- Damage Source
	SetExpressionVariable(aTranslation,"PlayerMelDef",PlayerAttrib:GetMeleeDefence());
	SetExpressionVariable(aTranslation,"PlayerRngDef",PlayerAttrib:GetRangeDefence());
	SetExpressionVariable(aTranslation,"PlayerTacDef",PlayerAttrib:GetTacticalDefence());
	-- Damage Type
	SetExpressionVariable(aTranslation,"PlayerPhyMit",PlayerAttrib:GetCommonMitigation());
	SetExpressionVariable(aTranslation,"PlayerOFMit",PlayerAttrib:GetPhysicalMitigation());
	SetExpressionVariable(aTranslation,"PlayerTacMit",PlayerAttrib:GetTacticalMitigation());
	SetExpressionVariable(aTranslation,"PlayerFireMit",PlayerAttrib:GetFireMitigation());
	SetExpressionVariable(aTranslation,"PlayerLightningMit",PlayerAttrib:GetLightningMitigation());
	SetExpressionVariable(aTranslation,"PlayerFrostMit",PlayerAttrib:GetFrostMitigation());
	SetExpressionVariable(aTranslation,"PlayerAcidMit",PlayerAttrib:GetAcidMitigation());
	SetExpressionVariable(aTranslation,"PlayerShadowMit",PlayerAttrib:GetShadowMitigation());
end

local function ExecuteMyExpression()
	-- clear result output
	ltbOutput:SetText("");
	local sExpression = ltbExpression:GetText();
	
	-- process escape characters
	sExpression = sExpression:gsub("\\\\", "&backslash&");
	sExpression = sExpression:gsub("\\n", "\n");
	--sExpression = sExpression:gsub("\\t", "\t"); --not supported by Lotro
	sExpression = sExpression:gsub("&backslash&", "\\");
	
	-- get variables
	local nL = tonumber(ltbLevel:GetText());
	if nL == nil then
		nL = 1;
	end
	local nToL = tonumber(ltbToLevel:GetText());
	if nToL == nil then
		nToL = nL;
	end
	local nLvlStep = 1;
	if nL > nToL then
		nLvlStep = -1;
	end
	local nN = tonumber(ltbNorC:GetText());
	local sC = "";
	if nN == nil then
		nN = 1;
		sC = ltbNorC:GetText();
	end
	
	local nErrCode;
	
	-- translate expression
	local aTranslation = TranslateExpression(sExpression,{{"L",nL},{"N",nN},{"C",sC}});
	
	SetPlayerVariables(aTranslation);
	
	-- execute if no error encountered during translation
	local nErrCode = GetExpressionError(aTranslation);
	if nErrCode == 0 then
		local sResult;
		for nCurrentL = nL, nToL, nLvlStep do
			-- set level
			SetExpressionVariable(aTranslation,"L",nCurrentL);
			
			-- execute expression
			sResult = ExecuteExpression(aTranslation);
			
			-- check for error
			nErrCode = GetExpressionError(aTranslation);
			if nErrCode ~= 0 then
				break; -- break off on error
			end
			
			-- output result
			ltbOutput:AppendText(sResult);
			if nL ~= nToL then
				ltbOutput:AppendText("\n");
			end
		end
	end
	
	-- display error info
	if nErrCode ~= 0 then
		local sErrMsg, sExprWIP;
		nErrCode, sErrMsg, sExprWIP = GetExpressionError(aTranslation);
		ltbOutput:AppendText("Error:"..nErrCode.." ("..sErrMsg..")");
		if sExprWIP:match("^%s*%[(%d+)%]%s*$") == nil then
			ltbOutput:AppendText("\nWIP:"..sExprWIP);
		end
	end
end

local function SetLayout()
	local nCtrlStartX = 25;
	local nCtrlStartY = 50;
	local nCtrlCursorX = nCtrlStartX;
	local nCtrlCursorY = nCtrlStartY;
	local nCtrlSpacingX = 45;
	local nCtrlSpacingY = 20;

	local nExprOutpWidth = wndMain:GetWidth()-2*nCtrlStartX;
	local nExprWidth = nExprOutpWidth;
	local nOutpWidth = nExprOutpWidth;
	local nTotalExprOutpHeight = wndMain:GetHeight()-242;
	local nExprHeight = 0.35*nTotalExprOutpHeight;
	local nOutpHeight = nTotalExprOutpHeight-nExprHeight;

	lblLevel:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lblLevel:GetWidth();
	ltbLevel:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+ltbLevel:GetWidth();
	lblToLevel:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lblToLevel:GetWidth();
	ltbToLevel:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+ltbToLevel:GetWidth()+nCtrlSpacingX;
	lblNorC:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lblNorC:GetWidth();
	ltbNorC:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+ltbNorC:GetWidth();

	nCtrlCursorX = nCtrlStartX;
	nCtrlCursorY = nCtrlCursorY+ltbLevel:GetHeight()+nCtrlSpacingY;

	lblExpression:SetPosition(nCtrlCursorX,nCtrlCursorY);
	lblExpression:SetWidth(nExprWidth);
	nCtrlCursorY = nCtrlCursorY+lblExpression:GetHeight();
	ltbExpression:SetPosition(nCtrlCursorX,nCtrlCursorY);
	ltbExpression:SetSize(nExprWidth,nExprHeight);
	nCtrlCursorY = nCtrlCursorY+ltbExpression:GetHeight()+nCtrlSpacingY;

	lbtnExecute:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lbtnExecute:GetWidth()+nCtrlSpacingX;
	lbtnClear:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lbtnClear:GetWidth()+nCtrlSpacingX;
	lbtnHelp:SetPosition(nCtrlCursorX,nCtrlCursorY);
	nCtrlCursorX = nCtrlCursorX+lbtnHelp:GetWidth()+nCtrlSpacingX;
	
	nCtrlCursorX = nCtrlStartX;
	nCtrlCursorY = nCtrlCursorY+lbtnExecute:GetHeight()+nCtrlSpacingY;

	lblOutput:SetPosition(nCtrlCursorX,nCtrlCursorY);
	lblOutput:SetWidth(nOutpWidth);
	nCtrlCursorY = nCtrlCursorY+lblOutput:GetHeight();
	ltbOutput:SetPosition(nCtrlCursorX,nCtrlCursorY);
	ltbOutput:SetSize(nOutpWidth,nOutpHeight);
	nCtrlCursorY = nCtrlCursorY+ltbOutput:GetHeight()+nCtrlSpacingY;
end

Player = Turbine.Gameplay.LocalPlayer.GetInstance();
PlayerAttrib = Player:GetAttributes();

-- create main window + controls
wndMain = Turbine.UI.Lotro.Window();
wndMain:SetSize(850,750);
local displayWidth, displayHeight = Turbine.UI.Display.GetSize();
local windowWidth, windowHeight = wndMain:GetSize();
wndMain:SetPosition((displayWidth-windowWidth)/2,(displayHeight-windowHeight)/2);
wndMain:SetText("CalcStat Expressions");

local nLabelFont = Turbine.UI.Lotro.Font.Verdana18;
local nLabelHeight = 22;
local nTextBoxFont = Turbine.UI.Lotro.Font.Verdana18;
local nTextBoxHeight = 22;

lblLevel = Turbine.UI.Label();
lblLevel:SetSize(75,nLabelHeight);
lblLevel:SetFont(nLabelFont);
lblLevel:SetText("$L(evel):");
lblLevel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
lblLevel:SetParent(wndMain);
lblLevel:SetVisible(true);
ltbLevel = Turbine.UI.Lotro.TextBox();
ltbLevel:SetSize(40,nTextBoxHeight);
ltbLevel:SetFont(nTextBoxFont);
ltbLevel:SetMultiline(false);
ltbLevel:SetParent(wndMain);
ltbLevel:SetVisible(true);
function lblLevel:MouseClick(sender, args) ltbLevel:Focus() end;
lblToLevel = Turbine.UI.Label();
lblToLevel:SetSize(40,nLabelHeight);
lblToLevel:SetFont(nLabelFont);
lblToLevel:SetText("to");
lblToLevel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
lblToLevel:SetParent(wndMain);
lblToLevel:SetVisible(true);
ltbToLevel = Turbine.UI.Lotro.TextBox();
ltbToLevel:SetSize(40,nTextBoxHeight);
ltbToLevel:SetFont(nTextBoxFont);
ltbToLevel:SetMultiline(false);
ltbToLevel:SetParent(wndMain);
ltbToLevel:SetVisible(true);
function lblToLevel:MouseClick(sender, args) ltbToLevel:Focus() end;

lblNorC = Turbine.UI.Label();
lblNorC:SetSize(82,nLabelHeight);
lblNorC:SetFont(nLabelFont);
lblNorC:SetText("$N or $C:");
lblNorC:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
lblNorC:SetParent(wndMain);
lblNorC:SetVisible(true);
ltbNorC = Turbine.UI.Lotro.TextBox();
ltbNorC:SetSize(200,nTextBoxHeight);
ltbNorC:SetFont(nTextBoxFont);
ltbNorC:SetMultiline(false);
ltbNorC:SetParent(wndMain);
ltbNorC:SetVisible(true);
function lblNorC:MouseClick(sender, args) ltbNorC:Focus() end;

lblExpression = Turbine.UI.Label();
lblExpression:SetHeight(nLabelHeight+6);
lblExpression:SetFont(nLabelFont);
lblExpression:SetBackColor(Turbine.UI.Color(0.1,0.15,0.35));
lblExpression:SetText("Enter Expression");
lblExpression:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
lblExpression:SetParent(wndMain);
lblExpression:SetVisible(true);
ltbExpression = Turbine.UI.Lotro.TextBox();
--ltbExpression:SetText("\"Race: \"..$PlayerRaceName..\"\\n\"..\n\"Class: \"..$PlayerClassName..\"\\n\"..\n\"Level: \"..$PlayerLevel..\"\\n\"..\n\"Outgoing Healing Rating: \"..$PlayerOutHeal..\"\\n\"..\n\"Outgoing Healing Percentage(\"..$PlayerOutHeal..\"): \"..\n\nWith($PercName = \"OutHeal\", $PercRating = $PlayerOutHeal:\nFormat(Afd(CalcStat($PercName..\"PRatP\",$PlayerLevel,$PercRating)+CalcStat($PercName..\"PBonus\",$PlayerLevel)),\"%.1f%%\"))");
ltbExpression:SetFont(nTextBoxFont);
ltbExpression:SetParent(wndMain);
ltbExpression:SetVisible(true);
function lblExpression:MouseClick(sender, args) ltbExpression:Focus() end;

lbtnExecute = Turbine.UI.Lotro.Button();
lbtnExecute:SetWidth(120);
lbtnExecute:SetText("Run Expression");
lbtnExecute:SetParent(wndMain);
lbtnExecute:SetVisible(true);
function lbtnExecute:MouseClick(sender, args) ExecuteMyExpression() end;

lbtnClear = Turbine.UI.Lotro.Button();
lbtnClear:SetWidth(100);
lbtnClear:SetText("Clear");
lbtnClear:SetParent(wndMain);
lbtnClear:SetVisible(true);
function lbtnClear:MouseClick(sender, args) ClearMyExpression() end;

lbtnHelp = Turbine.UI.Lotro.Button();
lbtnHelp:SetWidth(100);
lbtnHelp:SetText("Help");
lbtnHelp:SetParent(wndMain);
lbtnHelp:SetVisible(true);
function lbtnHelp:MouseClick(sender, args) ShowHelpOutput() end;

lblOutput = Turbine.UI.Label();
lblOutput:SetHeight(nLabelHeight+6);
lblOutput:SetFont(nLabelFont);
lblOutput:SetBackColor(Turbine.UI.Color(0.1,0.15,0.35));
lblOutput:SetText("Result");
lblOutput:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
lblOutput:SetParent(wndMain);
lblOutput:SetVisible(true);
ltbOutput = Turbine.UI.Lotro.TextBox();
ltbOutput:SetFont(nTextBoxFont);
ltbOutput:SetParent(wndMain);
ltbOutput:SetVisible(true);

wndMain:SetResizable(true);
function wndMain:SizeChanged(sender, args) SetLayout() end;

wndMain:SetVisible(true);

cmdCalcStat = Turbine.ShellCommand();
function cmdCalcStat:Execute(command, args)
	if type(args) == "string" then
		local sArgs = args:upper();
		if sArgs == "SHOW" then
			wndMain:SetVisible(true);
		elseif sArgs == "HIDE" then
			wndMain:SetVisible(false);
		else
			Turbine.Shell.WriteLine("no valid option provided");
			Turbine.Shell.WriteLine("usage: /calcstat show|hide");
		end
	end
end
Turbine.Shell.AddCommand("calcstat",cmdCalcStat);

Turbine.Shell.WriteLine("CalcStat version "..CalcStat("-Version",0).." loaded.");
Turbine.Shell.WriteLine("usage: /calcstat show|hide");
